#include "msr-index.h"
#include "emu64.h"

/* --------------------------------------------------------------------- */
/* inline asm bits                                                       */

static inline void vmexit_panic(void)
{
    asm volatile("outb %%al,$0xeb"
		 : : : "memory");
}

/* --------------------------------------------------------------------- */
/* helper functions                                                      */

int panic(char *message)
{
    printk("panic: %s\n", message);
    vmexit_panic();
    return 0;
}

int bounce_trap(struct regs_64 *regs, int trapno, int cbno)
{
    uint64_t *stack, rip, rsp, stack_cs, stack_rflags;
    int error_code = 0;
    int interrupt = 0;
    int failsafe = 0;
    int k = 0;

    if (trapno >= 0) {
	/* trap bounce */
	rip  = xen_ctxt.trap_ctxt[trapno].address;
	if (TI_GET_IF(&xen_ctxt.trap_ctxt[trapno]))
	    interrupt = 1;
	switch (trapno) {
	case 14: /* page fault */
	    shared_info.vcpu_info[0].arch.cr2 = read_cr2();
	case 13: /* general protection */
	    error_code = 1;
	    break;
	}
    }
    if (cbno >= 0) {
	/* callback */
	rip  = xencb[cbno];
	switch (cbno) {
	case CALLBACKTYPE_event:
 	    interrupt = 1;
	    break;
	}
    }

    if (!rip)
	panic("no guest trap handler");

    /* set interrupt flag depending on event channel mask */
    stack_rflags = regs->rflags & ~X86_EFLAGS_IF;
    if (!shared_info.vcpu_info[0].evtchn_upcall_mask)
	stack_rflags |= X86_EFLAGS_IF;

    /* old evtchn_upcall_mask is saved in cs slot on the stack */
    stack_cs = regs->cs | ((uint64_t)shared_info.vcpu_info[0].evtchn_upcall_mask << 32);
    if (interrupt)
	shared_info.vcpu_info[0].evtchn_upcall_mask = 1;

    if (!is_kernel()) {
	/* user mode */
	switch_mode();
	rsp = xen_ctxt.kernel_sp;
    } else {
	/* kernel mode */
	stack_cs &= ~3;         /* signal kernel mode */
	rsp = regs->rsp & ~0xf; /* align stack */
    }
    stack = (void*)(rsp);

#if 0
    if ((stack_cs & 3)  ||  cbno == CALLBACKTYPE_syscall) /* from user mode */
	printk("%s: code %" PRIx64 " stack %" PRIx64 "\n",
	       __FUNCTION__, rip, rsp);
#endif
    
    stack[-(++k)] = regs->ss;        // push ss
    stack[-(++k)] = regs->rsp;       // push rsp
    stack[-(++k)] = stack_rflags;    // push rflags
    stack[-(++k)] = stack_cs;        // push cs
    stack[-(++k)] = regs->rip;       // push rip
    if (error_code)
	stack[-(++k)] = regs->error; // push error code

    if (failsafe) {
	/* push segment registers */;
    }

    stack[-(++k)] = regs->r11;       // push r11
    stack[-(++k)] = regs->rcx;       // push rcx

    /* prepare emu stack, so iret jumps to the kernels handler. */
    regs->rip     = rip;
    regs->cs      = FLAT_KERNEL_CS;
    regs->rflags &= EFLAGS_TRAPMASK;
    regs->rsp     = rsp - 8*k;
    regs->ss      = FLAT_KERNEL_SS;

#if 0
    if (cbno == CALLBACKTYPE_syscall)
	regs->rflags |= X86_EFLAGS_TF;
#endif

    return 0;
}

static void print_gpf_info(struct regs_64 *regs)
{
    printk("%s: index 0x%x%s%s%s, rip %lx\n", __FUNCTION__,
	   (int)(regs->error >> 3),
	   (regs->error & 0x04) ? ", TI"  : "",
	   (regs->error & 0x02) ? ", IDT" : "",
	   (regs->error & 0x01) ? ", EXT" : "",
	   regs->rip);
}

static void setup_idt(void)
{
    int i;

    for (i = 0; i < 256; i++)
	xen_idt[i] = mkgate64(0xe008, (intptr_t)int_unknown, 0x8e);
    
    xen_idt[    0 ] = mkgate64(0xe008, (intptr_t)division_by_zero,    0x8e);
    xen_idt[    1 ] = mkgate64(0xe008, (intptr_t)debug_trap,          0x8e);
    xen_idt[    3 ] = mkgate64(0xe008, (intptr_t)debug_int3,          0xee);
    xen_idt[    6 ] = mkgate64(0xe008, (intptr_t)illegal_instruction, 0x8e);
    xen_idt[   13 ] = mkgate64(0xe008, (intptr_t)general_protection,  0x8e);
    xen_idt[   14 ] = mkgate64(0xe008, (intptr_t)page_fault,          0x8e);

    xen_idt[ 0x20 ] = mkgate64(0xe008, (intptr_t)event_callback,      0x8e);
    xen_idt[ 0x80 ] = mkgate64(0xe008, (intptr_t)int_80,              0xee);
}

/* --------------------------------------------------------------------- */

static void print_page_fault_info(struct regs_64 *regs)
{
    printk("%s:%s%s%s%s%s, rip %lx, cr2 %lx, %s\n", __FUNCTION__,
	   regs->error & 0x01 ? " preset"  : " nopage",
	   regs->error & 0x02 ? " write"   : " read",
	   regs->error & 0x04 ? " user"    : " kernel",
	   regs->error & 0x08 ? " reserved-bit"  : "",
	   regs->error & 0x10 ? " instr-fetch"   : "",
	   regs->rip, read_cr2(),
	   is_kernel() ? "kernel-mode" : "user-mode");
}

/* --------------------------------------------------------------------- */
/* called from assembler                                                 */

asmlinkage void do_boot(struct regs_64 regs)
{
    setup_idt();
    
    /* syscall setup */
    wrmsrl(MSR_STAR, (((uint64_t)0xe023 << 48) |
		      ((uint64_t)0xe008 << 32)));
    wrmsrl(MSR_LSTAR, (intptr_t)syscall);
    wrmsrl(MSR_SYSCALL_MASK, X86_EFLAGS_VM | X86_EFLAGS_RF |
	   X86_EFLAGS_NT | X86_EFLAGS_DF | X86_EFLAGS_IF | X86_EFLAGS_TF);
    
    /* register setup */
    regs.rax    = xen_ctxt.user_regs.rax;
    regs.rbx    = xen_ctxt.user_regs.rbx;
    regs.rcx    = xen_ctxt.user_regs.rcx;
    regs.rdx    = xen_ctxt.user_regs.rdx;
    regs.rsi    = xen_ctxt.user_regs.rsi;
    regs.rdi    = xen_ctxt.user_regs.rdi;
    regs.rbp    = xen_ctxt.user_regs.rbp;
    regs.rip    = xen_ctxt.user_regs.rip;
    regs.cs     = xen_ctxt.user_regs.cs;
    regs.rflags = xen_ctxt.user_regs.rflags;
    regs.rsp    = xen_ctxt.user_regs.rsp;
    regs.ss     = xen_ctxt.user_regs.ss;

#if 1
    asm volatile("mov %0, %%ds;\n" :: "r" (xen_ctxt.user_regs.ds) : "memory");
    asm volatile("mov %0, %%es;\n" :: "r" (xen_ctxt.user_regs.es) : "memory");
    asm volatile("mov %0, %%fs;\n" :: "r" (xen_ctxt.user_regs.fs) : "memory");
    asm volatile("mov %0, %%gs;\n" :: "r" (xen_ctxt.user_regs.gs) : "memory");
#endif

    /* jump to kernel (via iret in asm code calling us) */
    printk("booting kernel (entry %" PRIx64 ":%" PRIx64 ") ...\n",
	   regs.cs, regs.rip);
}

asmlinkage void do_illegal_instruction(struct regs_64 regs)
{
    int skip;

    emu_faults[XEN_FAULT_ILLEGAL_INSTRUCTION]++;
    if (context_is_emu(&regs)) {
	panic("ring0 (emu) illegal instruction");
    }
    if (context_is_user(&regs)) {
	uint8_t *i = (void*)regs.rip;
	printk("user ill: at %p"
	       "  0x%02x, 0x%02x, 0x%02x, 0x%02x,"
	       "  0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	       i, i[0], i[1], i[2], i[3], i[4], i[5], i[6], i[7]);
	bounce_trap(&regs, 6, -1);
	return;
    }
    
    skip = emulate_64(&regs);
    switch (skip) {
    case -1: /* error */
	panic("instruction emulation failed\n");
	break;
    case 0:  /* bounce to guest */
	bounce_trap(&regs, 6, -1);
	break;
    default: /* handled */
	regs.rip += skip;
	break;
    }
}

asmlinkage void do_general_protection(struct regs_64 regs)
{
    int skip;

    emu_faults[XEN_FAULT_GENERAL_PROTECTION]++;
    if (context_is_emu(&regs)) {
	print_gpf_info(&regs);
	panic("ring0 (emu) general protection fault");
    }
    if (context_is_user(&regs)) {
	emu_faults[XEN_FAULT_GENERAL_PROTECTION_GUEST]++;
	bounce_trap(&regs, 13, -1);
	return;
    }
    
    if (0 != regs.error) {
	print_gpf_info(&regs);
	panic("unhandled kernel gpf");
    }

    skip = emulate_64(&regs);
    switch (skip) {
    case -1: /* error */
	panic("instruction emulation failed\n");
	break;
    case 0:  /* bounce to guest */
	emu_faults[XEN_FAULT_GENERAL_PROTECTION_GUEST]++;
	bounce_trap(&regs, 13, -1);
	break;
    default: /* handled */
	emu_faults[XEN_FAULT_GENERAL_PROTECTION_EMUINS]++;
	regs.rip += skip;
	break;
    }
}

asmlinkage void do_page_fault(struct regs_64 regs)
{
    uint64_t cr2 = read_cr2();
    
    emu_faults[XEN_FAULT_PAGE_FAULT]++;
    if (context_is_emu(&regs)) {
	print_page_fault_info(&regs);
	panic("ring0 (emu) page fault");
    }

    if (context_is_user(&regs)) {
	if (regs.error & 0x01) {
	    /* present */
	    if (pgtable_fixup_flag(cr2, _PAGE_USER) > 0) {
		emu_faults[XEN_FAULT_PAGE_FAULT_FIX_USER]++;
		return;
	    }
	}
	goto bounce;
    }

    /* guest kernel page fault, fixup error code */
    regs.error &= ~0x04;

    if (cr2 < XEN_M2P_64)
	/* user space address */
	goto bounce;

    if ((regs.error & 0x03) == 0x03) {
	/* present, write */
	if (pgtable_fixup_flag(cr2, _PAGE_RW) > 0) {
	    emu_faults[XEN_FAULT_PAGE_FAULT_FIX_RO]++;
	    return;
	}
    }

    if (regs.error & 0x01) {
	/* present */
	if (pgtable_fixup_flag(cr2, _PAGE_USER) > 0) {
	    emu_faults[XEN_FAULT_PAGE_FAULT_FIX_USER]++;
	    return;
	}
    }
    
    print_page_fault_info(&regs);
    pgtable_walk(cr2, 0);
    panic("Oops");

bounce:
    emu_faults[XEN_FAULT_PAGE_FAULT_GUEST]++;
    bounce_trap(&regs, 14, -1);
}

asmlinkage void do_event_callback(struct regs_64 regs)
{
    if (shared_info.vcpu_info[0].evtchn_upcall_mask) {
	/* FIXME */
//	printk("%s: IRQs off\n", __FUNCTION__);
	return;
    }
    bounce_trap(&regs, -1, CALLBACKTYPE_event);
}

asmlinkage void do_int_80(struct regs_64 regs)
{
    emu_faults[XEN_FAULT_INT_80]++;
    bounce_trap(&regs, 0x80, -1);
}

asmlinkage void do_guest_forward(struct regs_64 regs)
{
    bounce_trap(&regs, regs.trapno, -1);
}
